# Sorting Algorithm Performance Analysis: Impact of Input Distribution on Runtime Performance

## Abstract

This study investigates how different sorting algorithms perform across various input distributions. Four algorithms (QuickSort, MergeSort, HeapSort, and InsertionSort) were evaluated using three input distributions (random, nearly sorted, and reverse sorted) with arrays of 10,000 elements. Performance was measured through execution time, comparison operations, and swap operations across 10 repetitions per configuration. Results show that QuickSort performs optimally on nearly sorted data, MergeSort excels with reverse sorted arrays, HeapSort demonstrates consistent performance regardless of distribution, and InsertionSort significantly underperforms compared to the other algorithms. These findings provide valuable insights for algorithm selection based on expected data characteristics in real-world applications.

## 1. Introduction

Sorting algorithms are fundamental components in computer science with wide-ranging applications in data processing, database management, and computational tasks. The performance of these algorithms varies significantly based on the characteristics of the input data. This research examines how different sorting algorithms respond to various input distributions, aiming to provide empirical data to guide algorithm selection in practical scenarios.

The research question addressed is: **How does the choice of sorting algorithm impact runtime performance across different input distributions (random, nearly sorted, reverse sorted)?**

We hypothesized that algorithms would perform differently depending on the input distribution, with some algorithms showing particular strengths or weaknesses with specific distribution types.

## 2. Methodology

### 2.1 Experiment Design

The experiment compared four well-known sorting algorithms:
- QuickSort
- MergeSort
- HeapSort
- InsertionSort

Each algorithm was tested against three input distributions:
- Random: Elements randomly distributed
- Nearly Sorted: Elements mostly in order with minimal displacement
- Reverse Sorted: Elements in descending order

### 2.2 Implementation Details

All algorithms were implemented in Python with instrumentation to measure:
- Execution time (seconds)
- Number of comparison operations
- Number of swap operations

For each algorithm and input distribution combination, 10 repetitions were conducted using arrays of 10,000 elements to ensure statistical reliability. The testing framework was designed to generate appropriate test data, execute the sorting algorithms, and collect performance metrics automatically.

### 2.3 Execution Progress

The experiment was conducted in phases:
1. **Control Group**: QuickSort with random input distribution
2. **Experimental Group - Partition 1**: QuickSort with nearly sorted and reverse sorted distributions; MergeSort with all three distribution types
3. **Experimental Group - Partition 2**: HeapSort with all three distribution types; InsertionSort with random and nearly sorted distributions
4. **Experimental Group - Partition 3**: InsertionSort with reverse sorted distribution (not completed due to time constraints)

### 2.4 Challenges Encountered

During experimentation, the following challenges were addressed:
- Ensuring consistent measurement methodology across all algorithms
- Managing large execution times for InsertionSort, which required significantly longer runtime
- Time constraints prevented the completion of InsertionSort testing on reverse sorted data

## 3. Results

### 3.1 Execution Time Performance

![Efficient algorithm execution time](efficient_algorithm_execution_time.png)
*<small>Fig 1: Execution time comparison between QuickSort, MergeSort, and HeapSort across different input distributions. QuickSort performs best on nearly sorted data, while MergeSort excels with reverse sorted data.</small>*

![Algorithm execution time log scale](all_algorithms_execution_time_log_scale.png)
*<small>Fig 2: Logarithmic scale comparison of execution times for all sorting algorithms. InsertionSort is orders of magnitude slower than the other algorithms, particularly for nearly sorted and random data.</small>*

The execution time data reveals several important patterns:

- QuickSort achieved the best performance on nearly sorted data (0.015043 seconds)
- MergeSort performed optimally on reverse sorted data (0.021612 seconds)
- HeapSort showed relatively consistent performance across all input distributions (ranging from 0.041724 to 0.044246 seconds)
- InsertionSort was dramatically slower than the other algorithms, with execution times of 3.663364 seconds for random data and 6.642791 seconds for nearly sorted data (approximately 150-300 times slower than the efficient algorithms)

### 3.2 Comparison Operations

![Comparison operations](comparison_operations.png)
*<small>Fig 3: Number of comparison operations for efficient sorting algorithms across different input distributions. MergeSort requires the fewest comparisons, especially on reverse sorted data, while HeapSort consistently requires the most comparisons.</small>*

The number of comparison operations varied significantly:

- MergeSort performed the fewest comparisons, particularly on reverse sorted data (64,608 comparisons)
- HeapSort consistently required the most comparisons across all distribution types (228,661 to 244,460)
- QuickSort showed variability based on input distribution, with nearly sorted data requiring fewer comparisons (139,693) compared to reverse sorted data (191,492)
- InsertionSort performed an extremely high number of comparisons (over 25 million for random data and over 44 million for nearly sorted data)

### 3.3 Swap Operations

![Swap operations](swap_operations.png)
*<small>Fig 4: Number of swap operations for efficient sorting algorithms across different input distributions. QuickSort requires fewer swaps for nearly sorted and reverse sorted data, while MergeSort maintains consistent swap counts regardless of input distribution.</small>*

Analysis of swap operations revealed:

- QuickSort required fewer swaps for non-random distributions (39,434 for nearly sorted and 32,309 for reverse sorted) compared to random data (71,003 to 73,700)
- MergeSort maintained a consistent number of swaps (133,616) regardless of input distribution
- HeapSort's swap operations varied based on distribution, with nearly sorted data requiring the fewest swaps (118,393) and reverse sorted data requiring the most (131,956)
- InsertionSort performed an extremely high number of swaps, matching its comparison count

### 3.4 QuickSort Performance Analysis

![QuickSort performance](quicksort_performance_analysis.png)
*<small>Fig 5: Detailed performance analysis of QuickSort across different input distributions. QuickSort performs best on nearly sorted data with the lowest execution time and requires fewer swaps for non-random distributions despite higher comparison counts for reverse sorted data.</small>*

QuickSort's performance varied noticeably across different input distributions:

- Best execution time: nearly sorted data (0.015043 seconds)
- Highest comparison count: reverse sorted data (191,492)
- Lowest swap operations: reverse sorted data (32,309)

This suggests that QuickSort's efficiency is not determined solely by comparison or swap counts but by a complex interplay between these operations and the specific distribution characteristics.

## 4. Conclusion and Future Work

### 4.1 Main Findings

1. **Algorithm Selection Matters**: The performance difference between sorting algorithms can be substantial, with execution times varying by orders of magnitude depending on the algorithm chosen.

2. **Distribution Sensitivity**: Each algorithm responds differently to input distributions:
   - QuickSort performs best on nearly sorted data
   - MergeSort excels with reverse sorted data
   - HeapSort shows consistent performance across distributions
   - InsertionSort is uniformly inefficient but appears particularly ill-suited for nearly sorted data

3. **Operation Tradeoffs**: The algorithms display different tradeoffs between comparison and swap operations. MergeSort minimizes comparisons but requires more swaps than QuickSort for certain distributions.

### 4.2 Practical Implications

These findings have important implications for algorithm selection in practical applications:

- For applications with predominantly nearly sorted data, QuickSort offers superior performance
- When dealing with potentially reverse sorted data, MergeSort is the optimal choice
- If distribution characteristics are unknown or highly variable, HeapSort provides the most consistent performance
- InsertionSort should generally be avoided for large datasets regardless of distribution

### 4.3 Future Work

Several directions for future research emerge from this study:

1. **Expanded Distribution Types**: Testing with additional distribution types such as few unique values, alternating sequences, or periodic patterns

2. **Hybrid Algorithms**: Investigating hybrid approaches that switch between algorithms based on detected input characteristics

3. **Scaling Analysis**: Examining how the relative performance of these algorithms scales with increasing input sizes

4. **Complete InsertionSort Testing**: Completing the missing experiments for InsertionSort with reverse sorted data

5. **Memory Usage Analysis**: Extending the performance metrics to include memory consumption patterns across algorithms and distributions

## 5. Appendices

### 5.1 Raw Results

#### Control Group
- **QuickSort, Random, 10000 elements, 10 repetitions**
  - Average Execution Time: 0.021211 seconds (first run) / 0.022282 seconds (second run)
  - Average Comparisons: 153247.6 / 155574.7
  - Average Swaps: 71002.9 / 73700.3

#### Experimental Group - Partition 1
- **QuickSort, Nearly Sorted, 10000 elements, 10 repetitions**
  - Average Execution Time: 0.015043 seconds
  - Average Comparisons: 139692.9
  - Average Swaps: 39433.7

- **QuickSort, Reverse Sorted, 10000 elements, 10 repetitions**
  - Average Execution Time: 0.018793 seconds
  - Average Comparisons: 191492.0
  - Average Swaps: 32309.0

- **MergeSort, Random, 10000 elements, 10 repetitions**
  - Average Execution Time: 0.027171 seconds
  - Average Comparisons: 120446.5
  - Average Swaps: 133616.0

- **MergeSort, Nearly Sorted, 10000 elements, 10 repetitions**
  - Average Execution Time: 0.025972 seconds
  - Average Comparisons: 115539.7
  - Average Swaps: 133616.0

- **MergeSort, Reverse Sorted, 10000 elements, 10 repetitions**
  - Average Execution Time: 0.021612 seconds
  - Average Comparisons: 64608.0
  - Average Swaps: 133616.0

#### Experimental Group - Partition 2
- **HeapSort, Random, 10000 elements, 10 repetitions**
  - Average Execution Time: 0.042874 seconds
  - Average Comparisons: 235341.2
  - Average Swaps: 124190.8

- **HeapSort, Nearly Sorted, 10000 elements, 10 repetitions**
  - Average Execution Time: 0.041724 seconds
  - Average Comparisons: 228661.7
  - Average Swaps: 118393.2

- **HeapSort, Reverse Sorted, 10000 elements, 10 repetitions**
  - Average Execution Time: 0.044246 seconds
  - Average Comparisons: 244460.0
  - Average Swaps: 131956.0

- **InsertionSort, Random, 10000 elements, 10 repetitions**
  - Average Execution Time: 3.663364 seconds
  - Average Comparisons: 25005568.1
  - Average Swaps: 25005568.6

- **InsertionSort, Nearly Sorted, 10000 elements, 10 repetitions**
  - Average Execution Time: 6.642791 seconds
  - Average Comparisons: 44173183.1
  - Average Swaps: 44173199.8

#### Experimental Group - Partition 3
- **InsertionSort, Reverse Sorted, 10000 elements**
  - Results not available (experiment not completed due to time constraints)

### 5.2 Figure References
1. efficient_algorithm_execution_time.png
2. all_algorithms_execution_time_log_scale.png
3. comparison_operations.png
4. swap_operations.png
5. quicksort_performance_analysis.png